
#include <chrono>
#include <climits>
#include <string_view>
#include "sqlite_modern_cpp.h"
#include <vector>

static const sqlite::sqlite_config sharedConfig {
    sqlite::OpenFlags::READWRITE | sqlite::OpenFlags::CREATE |
    sqlite::OpenFlags::NOMUTEX, nullptr, sqlite::Encoding::UTF8
};

namespace sqlite {
    template<>
    struct has_sqlite_type<std::string_view, SQLITE_BLOB, void> : std::true_type {};

    inline std::string_view get_val_from_db(sqlite3_value *value, result_type<std::string_view>) {
        auto type = sqlite3_value_type(value);
        switch (type) {
            case SQLITE_BLOB:
                return {(const char *)sqlite3_value_blob(value), uint32_t(sqlite3_value_bytes(value))};
            case SQLITE3_TEXT:
                return {(const char *)sqlite3_value_text(value), uint32_t(sqlite3_value_bytes(value))};
            default:
                break;
        }
        return {};
    }
    inline std::string_view get_col_from_db(sqlite3_stmt* stmt, int inx, result_type<std::string_view>) {
        auto type = sqlite3_column_type(stmt, inx);
        switch (type) {
            case SQLITE_BLOB:
                return {(char const *)sqlite3_column_blob(stmt, inx), (uint32_t)sqlite3_column_bytes(stmt, inx)};
            case SQLITE3_TEXT:
                return {(char const *)sqlite3_column_text(stmt, inx), (uint32_t)sqlite3_column_bytes(stmt, inx)};
            default:
                break;
        }
        return {};
    }
}

int main()
{
    sqlite::database db("words.db", sharedConfig);

    db << "PRAGMA cache_size=-2000000"; // in KB
    db << "PRAGMA temp_store=MEMORY";

    db.define("LEVENSTEIN", [](std::string_view s1, std::string_view s2) -> int {
        const auto m = s1.size();
        const auto n = s2.size();

        if(m == 0) return n;
        if(n == 0) return m;

        std::vector<size_t> costs;
        costs.resize(n + 1);
        for (auto i = 0u; i < n + 1; ++i)
            costs[i] = i + 1;

        for (auto i = 0u; i < m; ++i) {
            costs[0] = i + 1;
            std::size_t corner = i;
            for (auto j = 0u; j < n; ++j) {
                auto upper = costs[j + 1];
                if(s1[i] == s2[j]) {
                    costs[j + 1] = corner;
                } else {
                    auto t = std::min(upper, corner);
                    costs[j + 1] = std::min(costs[j], t) + 1;
                }
                corner = upper;
            }
        }
        return costs[n];
    });

    try {
        db << "CREATE TABLE IF NOT EXISTS left (word BLOB NOT NULL PRIMARY KEY);";
        db << "CREATE TABLE IF NOT EXISTS right (word BLOB NOT NULL PRIMARI KEY);";

        auto now = std::chrono::system_clock::now();
#if 1
        auto levenstein = db << "SELECT r.word as rw, MIN(LEVENSTEIN(?, r.word)) FROM right r";
        db << "SELECT l.word FROM left l ORDER BY l.word"
            >> [&](std::string_view v) {
                levenstein << v >> [v](std::string_view s1, int leve) {
                    std::cout << v << " : " << s1 << " : " << leve << '\n';
                };
                levenstein++;
            };
#else
        db << "SELECT l.word, r.word, MIN(LEVENSTEIN(l.word, r.word)) as l "
              "FROM left l JOIN right r GROUP BY l.word HAVING l < 60"
              >> [&](std::string_view s1, std::string_view s2, int leve) {
                  std::cout << s1 << " : " << s2 << " : " << leve << '\n';
            };
#endif
        auto end = std::chrono::system_clock::now();
        std::cout << std::chrono::duration_cast<std::chrono::milliseconds>(end - now).count() << '\n';
    } catch(const sqlite::sqlite_exception& e) {
        std::cout << e.what() << " : " << e.get_sql() << '\n';
    }
}
